"""Metadata package for libvcs."""

from __future__ import annotations

__title__ = "libvcs"
__package_name__ = "libvcs"
__description__ = "Lite, typed, python utilities for Git, SVN, Mercurial, etc."
__version__ = "0.45.1"
__author__ = "Tony Narlock"
__github__ = "https://github.com/vcs-python/libvcs"
__docs__ = "https://libvcs.git-pull.com"
__tracker__ = "https://github.com/vcs-python/libvcs/issues"
__pypi__ = "https://pypi.org/project/libvcs/"
__email__ = "tony@git-pull.com"
__license__ = "MIT"
__copyright__ = "Copyright 2016- Tony Narlock"
